# -*- coding: utf-8 -*-

"""
Updates the proprietary new_duration_days of a project in both the "new" and legacy
archives

This is a rewrite of the original proprietary-new_duration setter
(commands.py) by Richard Falardeau <rfalarde@nrao.edu>

"""
import argparse
import logging
import sys
import warnings
from argparse import Namespace
from enum import Enum

from sqlalchemy import exc as sa_exc
from astropy.time import Time
from pymygdala import LogHandler, SendNRAOEvent
from schema import create_session, Project, ExecutionBlock
from support.logging import get_console_logger, LOG_MESSAGE_FORMATTER

from ._version import ___version___
from .project_finders import NewArchiveProjectFinder, LegacyProjectFinder

# pylint: disable=logging-format-interpolation

MAX_DURATION = 730

_APPLICATION_NAME = 'proprietary_setter'
_LOG = get_console_logger(_APPLICATION_NAME)
_MISSING_PROFILE = """unable to derive the capo profile from the environment, """ + \
                   """provide the capo profile through the -P argument or set the """ + \
                   """CAPO_PROFILE environment variable."""
_DISALLOWED_DURATION = f'invalid proprietary new_duration; value must be between 0 ' \
                       f'and {MAX_DURATION}'

_DESCRIPTION = """A tool for setting the proprietary duration of a project"""
_EPILOG = """Return values:
0: everything worked,
1: can't deduce which profile to use
2: invalid proprietary new_duration
3: DB update failed"""

class ProprietaryPeriodSetter:
    ''' Sets the proprietary duration of a project to a new value
        in both the "new" and legacy archive DBs.

    '''

    def __init__(self, usage:str, args: Namespace):
        try:
            self.new_duration_days = args.duration
        except AttributeError as err:
            _LOG.error(f'{err}')
            _LOG.error(_DISALLOWED_DURATION)
            sys.exit(2)
        if self.new_duration_days not in range(0, MAX_DURATION):
            _LOG.error(_DISALLOWED_DURATION)
            _LOG.error(usage)
            sys.exit(2)

        try:
            self.project = args.project
        except AttributeError:
            _LOG.error('Unable to determine project code')
            _LOG.error(usage)
            sys.exit(3)

        try:
            assert args.profile is not None
            self.profile = args.profile
        except (AttributeError, AssertionError):
            _LOG.error(_MISSING_PROFILE)
            sys.exit(1)

    def set_project_proprietary_state(self):
        ''' Where the magic happens:
            sets proprietary duration and status of the specified project
        '''
        project_code = self.project
        new_arch_proj = NewArchiveProjectFinder(self.profile) \
            .find_project(self.project)
        if new_arch_proj is None:
            _LOG.error(f'project not found in new archive: {project_code}')
            sys.exit(3)

        if not self.is_science_project(project_code):
            _LOG.error('This script can be used only on science projects.')
            sys.exit(3)

        legacy_proj = LegacyProjectFinder(self.profile).find_projects(self.project)

        with warnings.catch_warnings():
            # suppress SQLAlchemy warnings
            warnings.simplefilter("ignore", category=sa_exc.SAWarning)

            new_arch_session = create_session('SDM', profile=self.profile)
            legacy_session = create_session('LEGACY', profile=self.profile)

            _LOG.debug(f'>>> profile: {self.profile}')
            try:
                _LOG.debug('>>> updating new archive')
                new_updated = self.update_new_archive(new_arch_proj)
                _LOG.debug('>>> updating legacy archive')
                legacy_updated = self.update_legacy_archive(legacy_proj)
                if new_updated and legacy_updated:
                    new_arch_session.commit()
                    if 'prod' in self.profile:
                        legacy_session.commit()
                    else:
                        _LOG.info('(This is where we would have updated the '
                                  'legacy archive if we were in production.)')
                        legacy_session.rollback()
                    self.announce_update(legacy_proj[0].stoptime)

            except UpdateException as exc:
                _LOG.error(f'update failure: {exc}')
                new_arch_session.rollback()
                legacy_session.rollback()
                sys.exit(3)
            except Exception as exc:
                _LOG.error(f'>>> SOME OTHER KIND OF FAILURE: {exc}')
                new_arch_session.rollback()
                legacy_session.rollback()
                sys.exit(3)

            finally:
                new_arch_session.close()
                legacy_session.close()

        return {'new': new_arch_proj, 'legacy': legacy_proj}

    def is_science_project(self, project_code: str):
        ''' Is this project a "science" project (i.e., VLA/EVLA)? '''
        session = None
        try:
            # suppress SQLAlchemy warnings
            with warnings.catch_warnings():

                session = create_session('SDM', profile=self.profile)
                project = session.query(ExecutionBlock) \
                    .filter(ExecutionBlock.project_code == project_code) \
                    .filter(ExecutionBlock.telescope.in_(
                        [Telescope.EVLA.value, Telescope.VLA.value])) \
                    .first()
                return project is not None
        except Exception as exc:
            _LOG.error(f'Problem checking telescope for {project_code}: {exc}')
            sys.exit(3)
        finally:
            if session is not None:
                session.close()

    def update_new_archive(self, project: Project):
        ''' Update the "new" archive DB with new project duration '''
        try:
            project.proprietary_duration = self.new_duration_days
            return True
        except Exception as ex:
            _LOG.error(f'{ex}')
            sys.exit(3)

    def update_legacy_archive(self, projects: list):
        ''' Update the legacy archive DB with new project duration

            PROPRIETARY_DURATION: input new_duration_days
            PROPRIETARY: STOPTIME + input new_duration_days
            PROJECT_LOCK: if proprietary > today then 'LTIME' else 'UTIME'
        '''
        try:
            for project_instance in projects:
                project_instance.proprietary_duration = self.new_duration_days
                project_instance.proprietary = project_instance.stoptime + self.new_duration_days
                new_mjd_endtime = Time.now().mjd
                if project_instance.proprietary > new_mjd_endtime:
                    project_instance.project_lock = ProjectLock.LTIME.value
                else:
                    project_instance.project_lock = ProjectLock.UTIME.value

            return True
        except Exception as ex:
            _LOG.error(f'{ex}')
            sys.exit(3)

    def announce_update(self, stoptime):
        '''
        Set up a LogHandler to record the fact we just made a change to
        this project.

        We're adding it here, instead of earlier, because nothing we log earlier
        should be presented to anyone but the command line user and would
        only add useless clutter to our system logging.

        We only really want the completed task to make a record in our system.

        :return:
        '''

        profile = self.profile if 'prod' in self.profile else 'nmtest'
        try:
            broadcast = LogHandler(profile=profile,
                                   application=_APPLICATION_NAME)
            broadcast.setLevel(logging.DEBUG)
            broadcast.setFormatter(LOG_MESSAGE_FORMATTER)
            _LOG.addHandler(broadcast)
        except Exception as exc:
            _LOG.error(f'{exc}')
            raise

        _LOG.info(f'Attempting to update proprietary period for '
                  f'{self.project}.')
        if self.new_duration_days != 0:
            _LOG.info(f'Locking for {self.new_duration_days} days from {stoptime}')
        else:
            _LOG.info('Unlocking')

        event = {'logData': {'project_code': self.project,
                             'proprietary_duration': self.new_duration_days,
                             'ingestion_type': 'evla_sdm'
                             },
                 'message': 'proprietary period updated',
                 'request': 're-index please'}
        SendNRAOEvent(profile=profile,
                      application=_APPLICATION_NAME) \
            .send(routing_key='ingestion-complete.metadata', event=event)
        return 0


def main(**kwargs):
    r"""
    The main entry point for this script.  Builds the parser, checks params, gets a profile and then
    attempts to update the db.  If that succeeds, we kick off an amygdala even to tell the system
    to re-index the project so the archive will reflect the new status.
    :param kwargs: command line arguments to be passed to our parser builder
    :return: nothing, if we complete the system will exit normally, if not we'll set a system
    exit code.
    """
    parser = make_parser()
    try:
        args = parser.parse_args(**kwargs)
    except AttributeError as err:
        _LOG.error(f'{err}')
        msg = str(err)
        if 'duration' in msg:
            _LOG.error(_DISALLOWED_DURATION)
            parser.print_help()
            sys.exit(2)
        elif 'project' in msg:
            _LOG.error('Unable to determine project code')
            sys.exit(3)
        else:
            _LOG.error(f'{err}')
            sys.exit(1)

    if args.duration not in range(0, 730):
        _LOG.error(_DISALLOWED_DURATION)
        parser.print_help()
        sys.exit(2)

    try:
        ProprietaryPeriodSetter(parser.format_usage(), args) \
            .set_project_proprietary_state()
        return 0
    except UpdateException as update_exception:
        _LOG.error(update_exception)
        parser.print_help()
        sys.exit(3)
    except ValueError as err:
        _LOG.error(f'{err}')
        if str(err) == 'could not determine profile':
            parser.print_help()
            sys.exit(1)
    except KeyError as err:
        # invalid profile => can't connect to new-archive DB
        if err.args[0].startswith('METADATADATABASE'):
            _LOG.error(f'Update failed: {err}')
            parser.print_help()
            sys.exit(3)

def make_parser():
    r"""
    Build a command line parser to take the parameters for setting a projects
    proprietary period.

    :return result: an argparse object with the profile, project and
    duration in its namespace"""
    result = argparse.ArgumentParser(
        description=_DESCRIPTION.format(___version___),
        formatter_class=argparse.RawTextHelpFormatter,
        epilog=_EPILOG)
    result.add_argument('-P', '--profile', action='store',
                        help='profile name to use, e.g. nmtest, dsoc-test, or nmprod')
    result.add_argument('project', action='store',
                        help='project_code to update proprietary duration')
    result.add_argument('duration', action='store',
                        type=int,
                        help='an integer duration to apply to the project; '
                             '0 (immediately public) to 730 (private for two years from today)')
    return result

class UpdateException(Exception):
    """ Wrapper for exceptions caught during our attempt to update the
    proprietary period.

    Since we aren't passing the argparse object to helper functions,
    we want to catch some library exceptions in sqlalchemy and also use it
    for some error trapping and 'handling'

    """


class Telescope(Enum):
    ''' all values we can expect in execution_block.telescope '''
    GBT  = 'GBT'
    VLA  = 'VLA'
    ALMA = 'ALMA'
    EVLA = 'EVLA'
    VLBA = 'VLBA'

class ProjectLock(Enum):
    ''' valid values for project.project_lock in legacy archive '''
    PUBLIC = 'PUBLIC'
    LTIME  = 'LTIME'
    UTIME  = 'UTIME'


if __name__ == '__main__':
    main()
